/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.api;

import javax.annotation.Nullable;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_243;
import net.minecraft.class_2586;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_3225;

/** Optional interface for {@link AbstractPart} implementations which can have sub-parts targeted and broken instead of
 * the whole thing. */
public interface SubdividedPart<Sub> {

    Class<Sub> getSubpartKeyClass();

    /** Multipart version of {@link class_2248#method_9606(class_2680, class_1937, class_2338, class_1657)} */
    void onSubpartBreakStart(class_1657 player, Sub subpart);

    /** Multipart version of {@link class_2248#method_9594(class_2680, class_1657, class_1922, class_2338)} */
    float calcSubpartBreakingDelta(Sub subpart);

    /** Multipart version of {@link class_2248#method_9576(class_1937, class_2338, class_2680, class_1657)}.
     * 
     * @return True if this should prevent Block.onBreak from being called afterwards, false otherwise. */
    default boolean onSubpartBreak(class_1657 player, Sub subpart) {
        return false;
    }

    /** Called instead of {@link class_1937#method_8650(class_2338, boolean)} in
     * {@link class_3225#method_14266}.
     * <p>
     * Generally this should remove the subpart from this {@link AbstractPart}, or return false if the whole part needs
     * to be removed from it's container.
     * 
     * @return True if this should prevent {@link MultipartContainer#removePart(AbstractPart)} from being called, false
     *         otherwise. */
    boolean clearSubpart(Sub subpart);

    /** Multipart version of {@link class_2248#method_9585(class_1936, class_2338, class_2680)} */
    void onSubpartBroken(Sub subpart);

    /** Multipart version of
     * {@link class_2248#method_9556(class_1937, class_1657, class_2338, class_2680, class_2586, class_1799)}. */
    void afterSubpartBreak(class_1657 player, class_1799 tool, Sub subpart);

    /** Subpart equivalent to {@link AbstractPart#getDynamicShape(float, class_243)}. If the given part is invalid then this should
     * return {@link AbstractPart#getDynamicShape(float, class_243)}. */
    class_265 getSubpartDynamicShape(class_243 hitVec, Sub subpart, float partialTicks);

    default boolean spawnBreakingParticles(class_243 hitVec, Sub subpart, class_2350 side) {
        return false;
    }

    /** @param hitVec The exact hit position, relative the the world's origin. (So you need to subtract the position of
     *            this part to get a position between 0 and 1).
     * @return The targeted subpart, or null if the given position doesn't intersect with a subpart. */
    @Nullable
    Sub getTargetedSubpart(class_243 hitVec);

    @Environment(EnvType.CLIENT)
    void playHitSound(class_1657 player, Sub subpart);
}
