/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.impl;

import java.util.ArrayList;
import java.util.List;

import javax.annotation.Nullable;
import net.minecraft.class_1922;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2586;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2741;
import net.minecraft.class_3611;
import net.minecraft.class_3612;
import alexiil.mc.lib.multipart.api.MultipartContainer;
import alexiil.mc.lib.multipart.api.MultipartContainer.MultipartCreator;
import alexiil.mc.lib.multipart.api.MultipartContainer.PartOffer;
import alexiil.mc.lib.multipart.api.MultipartHolder;
import alexiil.mc.lib.multipart.api.MultipartUtil;
import alexiil.mc.lib.multipart.api.NativeMultipart;

/** Contains the backend for {@link MultipartUtil}. */
public final class MultipartUtilImpl {

    @Nullable
    public static PartContainer get(class_1937 world, class_2338 pos) {
        return get((class_1922) world, pos);
    }

    @Nullable
    public static PartContainer get(class_1922 view, class_2338 pos) {
        class_2586 be = view.method_8321(pos);
        if (be instanceof MultipartBlockEntity) {
            return ((MultipartBlockEntity) be).container;
        }
        return null;
    }

    @Nullable
    public static MultipartContainer turnIntoMultipart(class_1937 world, class_2338 pos) {
        // See if there's an existing multipart that doesn't need to be changed
        PartContainer existing = get(world, pos);
        if (existing != null) {
            return existing;
        }

        class_3611 fluid = world.method_8316(pos).method_15772();
        boolean hasWater = fluid == class_3612.field_15910;
        if (fluid != class_3612.field_15910 && fluid != class_3612.field_15906) {
            return null;
        }

        class_2680 state = world.method_8320(pos);
        NativeMultipart nativeBlock = getNative(world, pos);
        if (nativeBlock != null) {
            List<MultipartCreator> conversions = nativeBlock.getMultipartConversion(world, pos, state);
            if (conversions == null) {
                return null;
            }
            if (!conversions.isEmpty()) {
                PartOffer offer = offerAdder(world, pos, hasWater, conversions, null, false);
                if (offer == null) {
                    return null;
                }
                offer.apply();
                return offer.getHolder().getContainer();
            }
        }
        return null;
    }

    private static NativeMultipart getNative(class_1937 world, class_2338 pos) {
        class_2680 state = world.method_8320(pos);
        if (state.method_26204() instanceof NativeMultipart nativeBlock) {
            return nativeBlock;
        }
        return NativeMultipart.LOOKUP.find(world, pos, null);
    }

    @Nullable
    public static PartOffer offerNewPart(
        class_1937 world, class_2338 pos, MultipartCreator creator, boolean respectEntityBBs
    ) {
        // See if there's an existing multipart that we can add to
        PartContainer currentContainer = get(world, pos);
        if (currentContainer != null) {
            return currentContainer.offerNewPart(creator, respectEntityBBs);
        }

        class_3611 fluid = world.method_8316(pos).method_15772();
        boolean hasWater = fluid == class_3612.field_15910;
        if (fluid != class_3612.field_15910 && fluid != class_3612.field_15906) {
            return null;
        }

        class_2680 state = world.method_8320(pos);
        NativeMultipart nativeBlock = getNative(world, pos);
        if (nativeBlock != null) {
            List<MultipartCreator> conversions = nativeBlock.getMultipartConversion(world, pos, state);
            if (conversions == null) {
                return null;
            }
            if (!conversions.isEmpty()) {
                return offerAdder(world, pos, hasWater, conversions, creator, respectEntityBBs);
            }
        } else if (!state.method_26207().method_15800()) {
            return null;
        }

        MultipartBlockEntity be = new MultipartBlockEntity(pos, state);
        be.method_31662(world);
        PartContainer container = new PartContainer(be);
        PartHolder holder = new PartHolder(container, creator);
        class_265 shape = holder.part.getCollisionShape();

        if (respectEntityBBs && !shape.method_1110()) {
            if (!world.method_8611(null, shape.method_1096(pos.method_10263(), pos.method_10264(), pos.method_10260()))) {
                return null;
            }
        }
        return new PartOffer() {
            @Override
            public MultipartHolder getHolder() {
                return holder;
            }

            @Override
            public void apply() {
                class_2680 newState = LibMultiPart.BLOCK.method_9564();
                newState = newState.method_11657(class_2741.field_12508, hasWater);
                world.method_8501(pos, newState);
                MultipartBlockEntity newBe = (MultipartBlockEntity) world.method_8321(pos);
                assert newBe != null;
                newBe.container = container;
                container.blockEntity = newBe;
                container.addPartInternal(holder);
            }
        };
    }

    @Nullable
    private static PartOffer offerAdder(
        class_1937 world, class_2338 pos, boolean hasWater, List<MultipartCreator> existing, MultipartCreator creatorB,
        boolean respectEntityBBs
    ) {
        MultipartBlockEntity be = new MultipartBlockEntity(
            pos, LibMultiPart.BLOCK.method_9564().method_11657(class_2741.field_12508, hasWater)
        );
        be.method_31662(world);
        PartContainer container = new PartContainer(be);

        List<PartHolder> existingHolders = new ArrayList<>();
        for (MultipartCreator creator : existing) {
            PartHolder holder = new PartHolder(container, creator);
            existingHolders.add(holder);

            // Add the existing ones so that they can intercept the offered part
            container.parts.add(holder);

            class_265 shape = holder.part.getCollisionShape();
            if (respectEntityBBs && !shape.method_1110()) {
                if (!world.method_8611(null, shape.method_1096(pos.method_10263(), pos.method_10264(), pos.method_10260()))) {
                    return null;
                }
            }
        }

        PartHolder offeredHolder = creatorB == null ? null : new PartHolder(container, creatorB);

        if (offeredHolder != null) {
            class_265 shape = offeredHolder.part.getCollisionShape();
            if (respectEntityBBs && !shape.method_1110()) {
                if (!world.method_8611(null, shape.method_1096(pos.method_10263(), pos.method_10264(), pos.method_10260()))) {
                    return null;
                }
            }

            container.recalculateShape();

            if (!container.canAdd(offeredHolder, true)) {
                return null;
            }
        }

        if (offeredHolder == null && existingHolders.isEmpty()) {
            return null;
        }

        return new PartOffer() {
            @Override
            public MultipartHolder getHolder() {
                return offeredHolder != null ? offeredHolder : existingHolders.get(0);
            }

            @Override
            public void apply() {
                // Cleanup the temporary additions
                container.parts.clear();

                // Actually place the new multipart
                class_2680 newState = LibMultiPart.BLOCK.method_9564();
                newState = newState.method_11657(class_2741.field_12508, hasWater);
                world.method_8501(pos, newState);
                MultipartBlockEntity newBe = (MultipartBlockEntity) world.method_8321(pos);
                assert newBe != null;
                newBe.container = container;
                container.blockEntity = newBe;
                for (PartHolder holder : existingHolders) {
                    container.addPartInternal(holder);
                }
                if (offeredHolder != null) {
                    container.addPartInternal(offeredHolder);
                }
            }
        };
    }
}
