/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.impl.client.model;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Supplier;

import com.google.common.collect.ImmutableList;
import com.mojang.datafixers.util.Pair;

import net.fabricmc.fabric.api.renderer.v1.model.FabricBakedModel;
import net.fabricmc.fabric.api.renderer.v1.render.RenderContext;
import net.fabricmc.fabric.api.rendering.data.v1.RenderAttachedBlockView;
import net.minecraft.class_1058;
import net.minecraft.class_1087;
import net.minecraft.class_1088;
import net.minecraft.class_1100;
import net.minecraft.class_1799;
import net.minecraft.class_1920;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_310;
import net.minecraft.class_3665;
import net.minecraft.class_4730;
import net.minecraft.class_777;
import net.minecraft.class_806;
import net.minecraft.class_809;
import alexiil.mc.lib.multipart.api.AbstractPart;
import alexiil.mc.lib.multipart.api.render.PartModelBaker;
import alexiil.mc.lib.multipart.api.render.PartModelKey;
import alexiil.mc.lib.multipart.api.render.PartRenderContext;
import alexiil.mc.lib.multipart.api.render.PartStaticModelRegisterEvent;
import alexiil.mc.lib.multipart.impl.MultipartBlockEntity;
import alexiil.mc.lib.multipart.impl.PartContainer;
import alexiil.mc.lib.multipart.impl.TransientPartIdentifier;
import alexiil.mc.lib.multipart.impl.TransientPartIdentifier.IdAdditional;
import alexiil.mc.lib.multipart.impl.TransientPartIdentifier.IdSubPart;
import alexiil.mc.lib.multipart.impl.client.PartModelData;
import alexiil.mc.lib.multipart.mixin.api.IBlockRenderManagerMixin;
import alexiil.mc.lib.multipart.mixin.api.IClientPlayerInteractionManagerMixin;

public final class MultipartModel
    implements class_1087, FabricBakedModel, PartStaticModelRegisterEvent.StaticModelRenderer {

    private final Map<Class<? extends PartModelKey>, PartModelBaker<?>> bakers, resolved;
    private final Function<class_4730, class_1058> textureGetter;

    public MultipartModel(Function<class_4730, class_1058> textureGetter) {
        this.textureGetter = textureGetter;
        bakers = new HashMap<>();
        resolved = new HashMap<>();
        PartStaticModelRegisterEvent.EVENT.invoker().registerModels(this);
    }

    @Override
    public <P extends PartModelKey> void register(Class<P> clazz, PartModelBaker<P> renderer) {
        bakers.put(clazz, renderer);
        resolved.clear();
    }

    @Override
    public class_1058 getSprite(class_2960 atlasId, class_2960 spriteId) {
        return textureGetter.apply(new class_4730(atlasId, spriteId));
    }

    public <P extends PartModelKey> PartModelBaker<? super P> getBaker(Class<P> clazz) {
        PartModelBaker<?> baker = resolved.get(clazz);
        if (baker != null) {
            return (PartModelBaker<? super P>) baker;
        }

        Class<? super P> c = clazz;
        do {
            PartModelBaker<?> pb = bakers.get(c);
            if (pb != null) {
                resolved.put(clazz, pb);
                return (PartModelBaker<? super P>) pb;
            }
        } while ((c = c.getSuperclass()) != null);
        resolved.put(clazz, NoopBaker.INSTANCE);
        return NoopBaker.INSTANCE;
    }

    @Override
    public List<class_777> method_4707(class_2680 state, class_2350 dir, Random rand) {
        return class_310.method_1551().method_1554().method_4744().method_4707(state, dir, rand);
    }

    @Override
    public boolean method_4708() {
        return true;
    }

    @Override
    public boolean method_4712() {
        return false;
    }

    @Override
    public boolean method_24304() {
        return true;
    }

    @Override
    public boolean method_4713() {
        return false;
    }

    @Override
    public class_1058 method_4711() {
        return class_310.method_1551().method_1541().method_3351().method_3333().method_4744()
            .method_4711();
    }

    @Override
    public class_809 method_4709() {
        return class_809.field_4301;
    }

    @Override
    public class_806 method_4710() {
        return class_806.field_4292;
    }

    // FabricBakedModel

    @Override
    public boolean isVanillaAdapter() {
        return false;
    }

    @Override
    public void emitBlockQuads(
        class_1920 blockView, class_2680 state, class_2338 pos, Supplier<Random> randomSupplier,
        RenderContext context
    ) {
        class_2586 be = blockView.method_8321(pos);
        if (be instanceof MultipartBlockEntity) {
            PartContainer container = ((MultipartBlockEntity) be).getContainer();

            class_310 mc = class_310.method_1551();
            if (
                mc.method_18854() && ((IBlockRenderManagerMixin) mc.method_1541())
                    .libmultipart_isRenderingPartiallyBrokenBlocks()
            ) {
                IClientPlayerInteractionManagerMixin interactionManager
                    = (IClientPlayerInteractionManagerMixin) mc.field_1761;
                if (pos.equals(interactionManager.libmultipart_getCurrentBreakPosition())) {
                    Object partKey = interactionManager.libmultipart_getPartKey();
                    if (partKey instanceof TransientPartIdentifier) {
                        TransientPartIdentifier identifier = (TransientPartIdentifier) partKey;
                        AbstractPart part = identifier.part;
                        PartModelKey modelKey = part.getModelKey();
                        Object subPart = null;
                        if (identifier.extra instanceof IdSubPart<?>) {
                            subPart = ((IdSubPart<?>) identifier.extra).subpart;
                        }
                        BreakingPartRenderContext partContext
                            = new BreakingPartRenderContext(context, false, randomSupplier, part, subPart);
                        if (modelKey != null) {
                            emitQuads(modelKey, modelKey.getClass(), partContext);
                        }
                        if (identifier.extra instanceof IdAdditional) {
                            for (AbstractPart additional : ((IdAdditional) identifier.extra).additional) {
                                PartModelKey key2 = additional.getModelKey();
                                if (key2 != null) {
                                    emitQuads(key2, key2.getClass(), partContext);
                                }
                            }
                        }
                        return;
                    }
                }
            }

            PartRenderContext ctx = new NormalPartRenderContext(context, false, randomSupplier);
            ImmutableList<PartModelKey> keys = container.getPartModelKeys();
            if (blockView instanceof RenderAttachedBlockView) {
                Object data = ((RenderAttachedBlockView) blockView).getBlockEntityRenderAttachment(pos);
                if (data instanceof PartModelData) {
                    keys = ((PartModelData) data).keys;
                }
            }
            for (PartModelKey key : keys) {
                emitQuads(key, key.getClass(), ctx);
            }
        }
    }

    private <K extends PartModelKey> void emitQuads(PartModelKey key, Class<K> clazz, PartRenderContext context) {
        PartModelBaker<? super K> baker = getBaker(clazz);
        if (baker != null) {
            baker.emitQuads(clazz.cast(key), context);
        }
    }

    @Override
    public void emitItemQuads(class_1799 stack, Supplier<Random> randomSupplier, RenderContext context) {
        // This isn't used for item models.
    }

    enum NoopBaker implements PartModelBaker<PartModelKey> {
        INSTANCE;

        @Override
        public void emitQuads(PartModelKey key, PartRenderContext ctx) {
            // NO-OP
        }
    }

    public static final class Unbaked implements class_1100 {
        MultipartModel model;

        @Override
        public Collection<class_2960> method_4755() {
            return Collections.emptySet();
        }

        @Override
        public Collection<class_4730> method_4754(
            Function<class_2960, class_1100> unbakedModelGetter, Set<Pair<String, String>> unresolvedTextureReferences
        ) {
            return Collections.emptySet();
        }

        @Override
        public class_1087 method_4753(
            class_1088 loader, Function<class_4730, class_1058> textureGetter, class_3665 rotationContainer,
            class_2960 modelId
        ) {
            if (model == null) {
                model = new MultipartModel(textureGetter);
            }
            return model;
        }
    }
}
