/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.impl.client.render;

import alexiil.mc.lib.multipart.mixin.api.IBlockMultipart;
import net.fabricmc.fabric.api.client.rendering.v1.WorldRenderContext;
import net.fabricmc.fabric.api.client.rendering.v1.WorldRenderEvents;
import net.minecraft.class_1921;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_239;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_310;
import net.minecraft.class_3532;
import net.minecraft.class_4587;
import net.minecraft.class_4588;
import java.util.Objects;

/**
 * Renders custom block outlines for only the selected part in a multipart block.
 */
public final class MultipartOutlineRenderer implements WorldRenderEvents.BlockOutline {
    public static final MultipartOutlineRenderer INSTANCE = new MultipartOutlineRenderer();

    private MultipartOutlineRenderer() {
    }

    @Override
    public boolean onBlockOutline(WorldRenderContext worldRenderContext,
                                  WorldRenderContext.BlockOutlineContext blockOutlineContext) {

        // In theory, this should never be null by this point, but who knows what custom renderers could be doing.
        class_239 hit = class_310.method_1551().field_1765;
        if (hit != null && hit.method_17783() == class_239.class_240.field_1332) {
            class_2680 state = blockOutlineContext.blockState();
            class_2248 block = state.method_26204();

            if (block instanceof IBlockMultipart<?> multipart) {
                class_2338 blockPos = blockOutlineContext.blockPos();
                class_265 partShape =
                        multipart.getPartOutline(state, worldRenderContext.world(), blockPos, hit.method_17784());

                if (!partShape.method_1110()) {
                    class_4588 linesConsumer =
                            Objects.requireNonNull(worldRenderContext.consumers(), "consumers is null")
                                    .getBuffer(class_1921.method_23594());
                    drawShapeOutline(worldRenderContext.matrixStack(), linesConsumer, partShape,
                            blockPos.method_10263() - blockOutlineContext.cameraX(),
                            blockPos.method_10264() - blockOutlineContext.cameraY(),
                            blockPos.method_10260() - blockOutlineContext.cameraZ());

                    return false;
                }
            }
        }
        return true;
    }

    private static void drawShapeOutline(class_4587 matrices, class_4588 vertexConsumer, class_265 voxelShape,
                                         double offsetX, double offsetY, double offsetZ) {
        class_4587.class_4665 entry = matrices.method_23760();
        voxelShape.method_1104((startX, startY, startZ, endX, endY, endZ) -> {
            float nx = (float) (endX - startX);
            float ny = (float) (endY - startY);
            float nz = (float) (endZ - startZ);

            float t = class_3532.method_22858(nx * nx + ny * ny + nz * nz);

            nx *= t;
            ny *= t;
            nz *= t;

            vertexConsumer.method_22918(entry.method_23761(), (float) (startX + offsetX), (float) (startY + offsetY),
                            (float) (startZ + offsetZ))
                    .method_22915(0f, 0f, 0f, 0.4f)
                    .method_23763(entry.method_23762(), nx, ny, nz)
                    .method_1344();
            vertexConsumer.method_22918(entry.method_23761(), (float) (endX + offsetX), (float) (endY + offsetY),
                            (float) (endZ + offsetZ))
                    .method_22915(0f, 0f, 0f, 0.4f)
                    .method_23763(entry.method_23762(), nx, ny, nz)
                    .method_1344();
        });
    }
}
